﻿using Newtonsoft.Json.Linq;
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using Threema.MsgApi.Exceptions;

namespace Threema.MsgApi.Messages
{
	class FileMessage : ThreemaMessage
	{
		public const int TYPE_CODE = 0x17;

		private const string KEY_BLOB_ID = "b";
		private const string KEY_THUMBNAIL_BLOB_ID = "t";
		private const string KEY_ENCRYPTION_KEY = "k";
		private const string KEY_MIME_TYPE = "m";
		private const string KEY_FILE_NAME = "n";
		private const string KEY_FILE_SIZE = "s";
		private const string KEY_TYPE = "i";
		private const string KEY_RENDERING_TYPE = "j";
		private const string KEY_CAPTION = "d";

		private readonly byte[] blobId;
		private readonly byte[] encryptionKey;
		private readonly string mimeType;
		private readonly string fileName;
		private readonly int fileSize;
		private readonly byte[] thumbnailBlobId;
		private readonly int renderingType = 0;
		private readonly string caption;

		public FileMessage(byte[] blobId, byte[] encryptionKey, String mimeType, String fileName, int fileSize, byte[] thumbnailBlobId, int renderingType = 0, string caption = null)
		{
			this.blobId = blobId;
			this.encryptionKey = encryptionKey;
			this.mimeType = mimeType;
			this.fileName = fileName;
			this.fileSize = fileSize;
			this.thumbnailBlobId = thumbnailBlobId;
			this.renderingType = renderingType;
			this.caption = caption;
		}

		public byte[] BlobId
		{
			get { return this.blobId; }
		}

		public byte[] EncryptionKey
		{
			get { return this.encryptionKey; }
		}

		public string MimeType
		{
			get { return this.mimeType; }
		}

		public string FileName
		{
			get { return this.fileName; }
		}

		public int FileSize
		{
			get { return this.fileSize; }
		}

		public byte[] ThumbnailBlobId
		{
			get { return this.thumbnailBlobId; }
		}

		public int RenderingType
		{
			get { return this.renderingType; }
		}

		public string Caption
		{
			get { return this.caption; }
		}

		public override int GetTypeCode()
		{
			return TYPE_CODE;
		}

		public override string ToString()
		{
			return string.Format("file message: {0}\ncaption: {1}", this.fileName, this.caption);
		}

		public override byte[] GetData()
		{
			JObject jo = new JObject();
			try
			{
				jo.Add(KEY_BLOB_ID, JToken.FromObject(DataUtils.ByteArrayToHexString(this.blobId)));
				if (this.thumbnailBlobId != null)
				{
					jo.Add(KEY_THUMBNAIL_BLOB_ID, JToken.FromObject(DataUtils.ByteArrayToHexString(this.thumbnailBlobId)));
				}
				jo.Add(KEY_ENCRYPTION_KEY, JToken.FromObject(DataUtils.ByteArrayToHexString(this.encryptionKey)));
				jo.Add(KEY_MIME_TYPE, JToken.FromObject(this.mimeType));
				jo.Add(KEY_FILE_NAME, JToken.FromObject(this.fileName));
				jo.Add(KEY_FILE_SIZE, JToken.FromObject(this.fileSize));
				jo.Add(KEY_TYPE, JToken.FromObject(0));
				jo.Add(KEY_RENDERING_TYPE, JToken.FromObject(0));
			}
			catch (Exception)
			{
				throw new BadMessageException();
			}

			return Encoding.UTF8.GetBytes(jo.ToString());
		}

		public static FileMessage FromString(string json)
		{
			try
			{
				JObject jo = JObject.Parse(json);
				byte[] encryptionKey = DataUtils.HexStringToByteArray(jo[KEY_ENCRYPTION_KEY].Value<string>());
				string mimeType = jo[KEY_MIME_TYPE].Value<string>();
				int fileSize = jo[KEY_FILE_SIZE].Value<int>();
				byte[] blobId = DataUtils.HexStringToByteArray(jo[KEY_BLOB_ID].Value<string>());

				string fileName;
				byte[] thumbnailBlobId = null;
				int renderingType = 0;
				string caption = null;

				//optional field
				if (jo.Children().Any(e => e.Path.EndsWith(KEY_THUMBNAIL_BLOB_ID)))
				{
					thumbnailBlobId = DataUtils.HexStringToByteArray(jo[KEY_THUMBNAIL_BLOB_ID].Value<string>());
				}

				if (jo.Children().Any(e => e.Path.EndsWith(KEY_FILE_NAME)))
				{
					fileName = jo[KEY_FILE_NAME].Value<string>();
				}
				else
				{
					fileName = "unnamed";
				}

				if (jo.Children().Any(e => e.Path.EndsWith(KEY_RENDERING_TYPE)))
				{
					renderingType = jo[KEY_RENDERING_TYPE].Value<int>();
				}
				else if (jo.Children().Any(e => e.Path.EndsWith(KEY_TYPE)))
				{
					renderingType = jo[KEY_TYPE].Value<int>();
				}
				else
				{
					renderingType = 0;
				}

				if (jo.Children().Any(e => e.Path.EndsWith(KEY_CAPTION)))
				{
					caption = jo[KEY_CAPTION].Value<string>();
				}

				return new FileMessage(
						blobId,
						encryptionKey,
						mimeType,
						fileName,
						fileSize,
						thumbnailBlobId,
						renderingType,
						caption
				);
			}
			catch
			{
				throw new BadMessageException();
			}
		}
	}
}
